from collections import namedtuple
from enum import Enum



class ProvidersEnum(Enum):
    OPENAI = "openai"
    CHAT_OPENAI = "chat_openai"
    ANTHROPIC = "anthropic"
    CHAT_ANTHROPIC = "chat_anthropic"
    AZURE_CHAT_OPENAI = "azure_chat_openai"
    COHERE = "cohere"
    DALLE = "dall-e"
    TOGETHER_AI = "together_ai"
    MISTRAL_AI = "mistral_ai"
    PERPLEXITY = "perplexity"
    STABLE_DIFFUSION = "stable-diffusion"
    DALL_E = "dall-e"
    OPENROUTER = "openrouter"
    REPLICATE = "replicate"
    OPENAI_EMBEDDINGS = "openai_embeddings"
    AZURE_OPENAI_EMBEDDINGS = "azure_openai_embeddings"
    AWS_BEDROCK = "aws_bedrock"


class ModelsEnum(Enum):
    OPENAI_GPT_35_TURBO_16K = "gpt-3.5-turbo-16k"
    AZURE_GPT_35_TURBO_16K = "gpt-35-turbo-16k"

    OPENAI_GPT_35_TURBO_1106 = "gpt-3.5-turbo-1106"

    OPENAI_GPT_35_TURBO_0613 = "gpt-3.5-turbo-0613"

    AZURE_GPT_35_TURBO = "gpt-35-turbo"

    OPENAI_GPT_4 = "gpt-4"
    AZURE_GPT_4 = "gpt-4"

    OPENAI_GPT_4_0613 = "gpt-4-0613"

    OPENAI_GPT_4_32K = "gpt-4-32k"
    AZURE_GPT_4_32K = "gpt-4-32k"

    OPENAI_GPT_35_TURBO_INSTRUCT = "gpt-3.5-turbo-instruct"

    CLAUDE_INSTANT_12 = "claude-instant-1.2"
    CLAUDE_4_5_HAIKU = "claude-haiku-4-5-20251001"

    DALLE_3 = "dall-e-3"
    
    STABLE_DIFFUSION_XL = "stable-diffusion-xl-1024-v1-0"
    STABLE_DIFFUSION_V6 = "stable-diffusion-v1-6"

    REPLICATE_FLUX = "black-forest-labs/flux-dev"
    REPLICATE_FLUX_PRO = "black-forest-labs/flux-pro"
    REPLICATE_FLUX_11_PRO = "black-forest-labs/flux-1.1-pro"
    REPLICATE_FLUX_SCHNELL = "black-forest-labs/flux-schnell"

    OPENAI_EMBEDDINGS_ADA_002 = "text-embedding-ada-002"
    AZURE_OPENAI_EMBEDDINGS_ADA_002 = "text-embedding-ada-002"
    OPENAI_EMBEDDINGS_TEXT_SMALL_3 = "text-embedding-3-small"

class StreamingEventType(Enum):
    UPDATE = "update"
    META = "meta"
    ERROR = "error"
    GENERATION_SOURCE = "generation_source"
    END = "end"


ModelDetails = namedtuple("ModelDetails", ["provider_name", "model_name"])


class GPTRouterEngines(Enum):
    OPENAI_GPT_35_TURBO = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-3.5-turbo"
    )
    OPENAI_GPT_35_TURBO_0613 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-3.5-turbo-0613"
    )
    OPENAI_GPT_35_TURBO_1106 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-3.5-turbo-1106"
    )
    OPENAI_GPT_35_TURBO_16K_0613 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-3.5-turbo-16k-0613"
    )
    OPENAI_GPT_35_TURBO_INSTRUCT = ModelDetails(
        ProvidersEnum.OPENAI.value, "gpt-3.5-turbo-instruct"
    )
    OPENAI_GPT_35_TURBO_0125 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-3.5-turbo-0125"
    )

    OPENAI_GPT_4_0613 = ModelDetails(ProvidersEnum.CHAT_OPENAI.value, "gpt-4-0613")
    OPENAI_GPT_4_TURBO = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-4-1106-preview"
    )
    OPENAI_GPT_4_TURBO_2024_04_09 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-4-turbo-2024-04-09"
    )

    OPENAI_GPT_4_TURBO_0125 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-4-0125-preview"
    )

    AZURE_OPENAI_GPT_4_TURBO_0125 = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt4-turbo-0125"
    )

    OPENAI_GPT_4_O = ModelDetails(ProvidersEnum.CHAT_OPENAI.value, "gpt-4o")
    
    OPENAI_GPT_4_O_MINI = ModelDetails(ProvidersEnum.CHAT_OPENAI.value, "gpt-4o-mini")
    AZURE_GPT_4O_MINI_2024_07_18 = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-4o-mini-2024-07-18"
    )
    OPENAI_GPT_4_O_24_08_06 = ModelDetails(ProvidersEnum.CHAT_OPENAI.value, "gpt-4o-2024-08-06")

    OPENAI_GPT_4_32K_0613 = ModelDetails(
        ProvidersEnum.CHAT_OPENAI.value, "gpt-4-32k-0613"
    )

    AZURE_GPT_35_TURBO = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-35-turbo"
    )
    AZURE_GPT_35_TURBO_16K = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-35-turbo-16k"
    )
    AZURE_GPT_35_TURBO_1106 = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-35-turbo-1106"
    )

    AZURE_GPT_4 = ModelDetails(ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-4")
    AZURE_GPT_4_TURBO = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-4-turbo"
    )
    AZURE_GPT_4_32K = ModelDetails(
        ProvidersEnum.AZURE_CHAT_OPENAI.value, "gpt-4-32k"
    )

    CLAUDE_INSTANT_12 = ModelDetails(
        ProvidersEnum.ANTHROPIC.value, "claude-instant-1.2"
    )
    CLAUDE_2 = ModelDetails(ProvidersEnum.ANTHROPIC.value, "claude-2")
    CLAUDE_21 = ModelDetails(ProvidersEnum.ANTHROPIC.value, "claude-2.1")
    CLAUDE_3_OPUS = ModelDetails(
        ProvidersEnum.ANTHROPIC.value, "claude-3-opus-20240229"
    )

    MISTRAL_7B = ModelDetails(
        ProvidersEnum.TOGETHER_AI.value, "mistralai/Mistral-7B-Instruct-v0.2"
    )
    MIXTRAL_8X7B = ModelDetails(
        ProvidersEnum.TOGETHER_AI.value, "mistralai/Mixtral-8x7B-Instruct-v0.1"
    )
    LLAMA_2_70B = ModelDetails(
        ProvidersEnum.TOGETHER_AI.value, "togethercomputer/llama-2-70b-chat"
    )

    MISTRAL_MEDIUM = ModelDetails(ProvidersEnum.MISTRAL_AI.value, "mistral-medium")

    PPTX_70B_ONLINE = ModelDetails(
        ProvidersEnum.PERPLEXITY.value, "pplx-70b-online"
    )
    PPTX_7B_ONLINE = ModelDetails(ProvidersEnum.PERPLEXITY.value, "pplx-7b-online")

    STABLE_DIFFUSION_XL = ModelDetails(
        ProvidersEnum.STABLE_DIFFUSION.value, "stable-diffusion-xl-1024-v1-0"
    )
    DALLE_3 = ModelDetails(ProvidersEnum.DALL_E.value, "dall-e-3")
    REPLICATE_FLUX = ModelDetails(
        ProvidersEnum.REPLICATE.value, "black-forest-labs/flux-dev"
    )
    REPLICATE_FLUX_PRO = ModelDetails(
        ProvidersEnum.REPLICATE.value, "black-forest-labs/flux-pro"
    )
    REPLICATE_FLUX_11_PRO = ModelDetails(
        ProvidersEnum.REPLICATE.value, "black-forest-labs/flux-1.1-pro"
    )
    REPLICATE_FLUX_SCHNELL = ModelDetails(
        ProvidersEnum.REPLICATE.value, "black-forest-labs/flux-schnell"
    )

    LLAMA_3_70B_NITRO = ModelDetails(
        ProvidersEnum.OPENROUTER, "meta-llama/llama-3-70b-instruct:nitro"
    )

    CLAUDE_3_SONNET = ModelDetails(
        ProvidersEnum.CHAT_ANTHROPIC.value, "claude-3-sonnet-20240229"
    )

    CLAUDE_3_HAIKU = ModelDetails(
        ProvidersEnum.CHAT_ANTHROPIC.value, "claude-3-haiku-20240307"
    )

    OPENROUTER_CLAUDE_3_HAIKU = ModelDetails(
        ProvidersEnum.OPENROUTER.value, "anthropic/claude-3-haiku"
    )

    CLAUDE_3_7_SONNET_2025_02_19 = ModelDetails(
        ProvidersEnum.CHAT_ANTHROPIC.value, "claude-3-7-sonnet-2025-02-19"
    )

    CLAUDE_4_5_HAIKU = ModelDetails(
        ProvidersEnum.CHAT_ANTHROPIC.value, "claude-haiku-4-5-20251001"
    )


from __future__ import annotations
import asyncio
from pydantic import TypeAdapter
import logging
import json
import httpx
from typing import Generator, List, Optional, Union

from gpt_router.models import (
    GPTRouterMetadata,
    ImageGenerationRequest,
    ImageGenerationResponse,
    ImageEditRequest,
    ModelGenerationRequest,
    GenerationResponse,
    ChunkedGenerationResponse,
    ModelRouterEmbeddingsGenerationRequest,
    ModelRouterGenerationResponse,
    ModelRouterEmbeddingsGenerationResponse,
)
from gpt_router.exceptions import (
    GPTRouterApiTimeoutError,
    GPTRouterBadRequestError,
    GPTRouterStreamingError,
    GPTRouterForbiddenError,
    GPTRouterInternalServerError,
    GPTRouterNotAvailableError,
    GPTRouterTooManyRequestsError,
    GPTRouterUnauthorizedError,
)
from gpt_router.constants import DEFAULT_REQUEST_TIMEOUT

from tenacity import (
    Retrying,
    AsyncRetrying,
    retry_if_exception_type,
    stop_after_attempt,
    wait_fixed,
)

logger = logging.getLogger(__name__)

STATUS_CODE_EXCEPTION_MAPPING = {
    400: GPTRouterBadRequestError,
    406: GPTRouterNotAvailableError,
    401: GPTRouterUnauthorizedError,
    403: GPTRouterForbiddenError,
    429: GPTRouterTooManyRequestsError,
    500: GPTRouterInternalServerError,
    503: GPTRouterNotAvailableError,
}


class ServerError(Exception):
    pass


class GPTRouterClient:
    models = None
    request_timeout = DEFAULT_REQUEST_TIMEOUT

    def __init__(
        self,
        base_url,
        api_key,
        request_timeout: int = 60,
        stream_read_timeout: Optional[int] = None,
        additional_metadata: Optional[GPTRouterMetadata] = None,
    ):
        self.base_url = base_url
        if api_key is None:
            raise ValueError("API key cannot be None. Please provide a valid API key.")
        self.api_key = api_key
        self.request_timeout = request_timeout
        self.additional_metadata = additional_metadata
        self.stream_read_timeout = (
            request_timeout if stream_read_timeout is None else stream_read_timeout
        )

    async def agenerate_embeddings(
        self,
        *,
        ordered_generation_requests: List[ModelRouterEmbeddingsGenerationRequest],
        model_router_metadata: Optional[GPTRouterMetadata] = None,
    ) -> ModelRouterGenerationResponse:
        api_payload = {
            "data": [
                {
                    **request.model_dump(),
                    "order": index + 1,
                }
                for index, request in enumerate(ordered_generation_requests)
            ],
        }
        api_payload = self.add_metadata_info(api_payload, model_router_metadata)
        response = await self._async_api_call(
            path="/v1/generate/embeddings",
            method="POST",
            payload=api_payload,
        )
        return ModelRouterEmbeddingsGenerationResponse.model_validate(response)

    async def agenerate_embeddings_v2(
        self,
        *,
        input: List[str],
        model_name: str,
        model_router_metadata: Optional[GPTRouterMetadata] = None,
    ) -> ModelRouterEmbeddingsGenerationResponse:
        """
        Generate embeddings using the v2 endpoint which uses round-robin load balancing
        across multiple providers for improved availability and performance.
        """
        api_payload = {
            "input": input,
            "modelName": model_name,
        }
        api_payload = self.add_metadata_info(api_payload, model_router_metadata)
        response = await self._async_api_call(
            path="/v1/generate/embeddings-v2",
            method="POST",
            payload=api_payload,
        )
        return ModelRouterEmbeddingsGenerationResponse.model_validate(response)

    def add_metadata_info(
        self,
        payload: dict,
        model_router_metadata: Optional[GPTRouterMetadata] = None,
    ):
        metadata = {}
        if self.additional_metadata:
            metadata.update(self.additional_metadata.model_dump())
        if model_router_metadata:
            metadata.update(model_router_metadata.model_dump())

        payload.update(
            {
                "metadata": metadata,
                "tag": metadata.get("tag"),
                "createdByUserId": metadata.get("created_by_user_id"),
                "historyId": (
                    str(metadata["history_id"]) if metadata.get("history_id") else None
                ),
                "appId": metadata.get("appId"),
            }
        )

        payload = {k: v for k, v in payload.items() if v is not None}
        return payload

    async def _async_api_call(self, *, path: str, method: str, payload: dict, **kwargs):
        try:
            async with httpx.AsyncClient() as client:
                response = await client.request(
                    method,
                    url=self.base_url.rstrip("/")
                    + ("/api" if not self.base_url.endswith("/api") else "")
                    + path,
                    headers={
                        "content-type": "application/json",
                        "ws-secret": self.api_key,
                    },
                    json=payload,
                    timeout=self.request_timeout,
                )
                if response.status_code == 200:
                    return response.json()
                elif response.status_code == 202 or response.status_code == 204:
                    return None
                else:
                    error_class = STATUS_CODE_EXCEPTION_MAPPING.get(
                        response.status_code, Exception
                    )
                    raise error_class(response.json())
        except httpx.TimeoutException as err:
            logger.error(f"Timeout error: {err}")
            raise GPTRouterApiTimeoutError("Api Request timed out")

    def _api_call(self, *, path: str, method: str, payload: dict, **kwargs):
        try:
            with httpx.Client() as client:
                response = client.request(
                    method,
                    url=self.base_url.rstrip("/")
                    + ("/api" if not self.base_url.endswith("/api") else "")
                    + path,
                    headers={
                        "content-type": "application/json",
                        "ws-secret": self.api_key,
                    },
                    json=payload,
                    timeout=self.request_timeout,
                )
                if response.status_code == 200:
                    return response.json()
                elif response.status_code == 202 or response.status_code == 204:
                    return None
                else:
                    error_class = STATUS_CODE_EXCEPTION_MAPPING.get(
                        response.status_code, Exception
                    )
                    raise error_class(response.json())
        except httpx.TimeoutException as err:
            logger.error(f"Timeout error: {err}")
            raise GPTRouterApiTimeoutError("Api Request timed out")

    async def astream_events(self, *, path: str, method: str, payload: dict, **kwargs):
        retry_count = kwargs.get("stream_retry_count") or 1
        stop_stream_signal = kwargs.get("stop_stream_signal")
        switch_models_order_on_timeout = (
            kwargs.get("switch_models_order_on_timeout") or False
        )
        async_retrying = AsyncRetrying(
            stop=stop_after_attempt(retry_count),
            wait=wait_fixed(0.5),
            retry=(
                retry_if_exception_type(httpx.ReadTimeout)
                | retry_if_exception_type(ServerError)
                | retry_if_exception_type(httpx.ReadError)
            ),
            before_sleep=lambda retry_state: logger.warn(
                f"Read timeout. Retrying... (Attempt {retry_state.attempt_number} of {retry_count})"
            ),
        )
        attempt_number = 0
        async for attempt in async_retrying:
            attempt_number += 1

            if attempt_number < retry_count:
                read_timeout = self.stream_read_timeout
            else:
                read_timeout = self.request_timeout
            try:
                async with httpx.AsyncClient() as client:
                    async with client.stream(
                        method,
                        url=self.base_url.rstrip("/")
                        + ("/api" if not self.base_url.endswith("/api") else "")
                        + path,
                        data=json.dumps(payload),
                        headers={
                            "Content-type": "application/json",
                            "ws-secret": self.api_key,
                        },
                        timeout=httpx.Timeout(self.request_timeout, read=read_timeout),
                    ) as response:
                        if response.status_code >= 500:
                            raise ServerError("Server error")
                        response.raise_for_status()
                        async for line in response.aiter_lines():
                            if stop_stream_signal and stop_stream_signal.is_set():
                                logger.info("Stopping stream due to signal")
                                await response.aclose()
                                return
                            try:
                                if line.strip() == "":
                                    continue

                                line_type, line_data = (
                                    segment.strip() for segment in line.split(":", 1)
                                )

                                if line_type != "data":
                                    continue

                                data: dict = json.loads(line_data.strip())
                                if data["event"] == "error":
                                    raise GPTRouterStreamingError(data)
                                yield TypeAdapter(ChunkedGenerationResponse).validate_python(data)
                            except GPTRouterStreamingError as e:
                                raise e
                            except Exception:
                                continue

                        return
            except (httpx.ReadTimeout, ServerError, httpx.ReadError) as err:
                logger.error(str(err))
                if attempt_number == retry_count:
                    logger.error("All retry attempts failed")
                    raise
                elif switch_models_order_on_timeout:
                    if attempt_number % 2:
                        self._reorder_models_on_timeout(payload)
                    logger.info(f"Retrying as per attempt #{attempt_number}")
            except httpx.HTTPStatusError as err:
                raise Exception(
                    f"HTTP Error {err.response.status_code}: {err.response.text}"
                ) from err
            except httpx.TimeoutException as err:
                raise TimeoutError("Request timed out") from err
            finally:
                if stop_stream_signal and stop_stream_signal.is_set():
                    logger.info("Stream stopped due to signal")
                    return

    def stream_events(
        self, *, path: str, method: str, payload: dict, **kwargs
    ) -> Generator[ChunkedGenerationResponse]:
        retry_count = kwargs.get("stream_retry_count") or 1
        switch_models_order_on_timeout = (
            kwargs.get("switch_models_order_on_timeout") or False
        )
        retrying = Retrying(
            stop=stop_after_attempt(retry_count),
            wait=wait_fixed(0.5),
            retry=(
                retry_if_exception_type(httpx.ReadTimeout)
                | retry_if_exception_type(ServerError)
                | retry_if_exception_type(httpx.ReadError)
            ),
            before_sleep=lambda retry_state: logger.warn(
                f"Read timeout. Retrying... (Attempt {retry_state.attempt_number} of {retry_count})"
            ),
        )

        attempt_number = 0
        for attempt in retrying:
            attempt_number += 1

            if attempt_number < retry_count:
                read_timeout = self.stream_read_timeout
            else:
                read_timeout = self.request_timeout

            try:
                with httpx.Client() as client:
                    with client.stream(
                        method=method,
                        url=self.base_url.rstrip("/")
                        + ("/api" if not self.base_url.endswith("/api") else "")
                        + path,
                        data=json.dumps(payload),
                        headers={
                            "Content-type": "application/json",
                            "ws-secret": self.api_key,
                        },
                        timeout=httpx.Timeout(self.request_timeout, read=read_timeout),
                    ) as response:
                        if response.status_code >= 500:
                            raise ServerError("Server error")
                        response.raise_for_status()
                        for line in response.iter_lines():
                            try:
                                if line.strip() == "":
                                    continue

                                line_type, line_data = (
                                    segment.strip() for segment in line.split(":", 1)
                                )
                                if line_type != "data":
                                    continue

                                data = json.loads(line_data.strip())
                                if data["event"].lower() == "error":
                                    raise GPTRouterStreamingError(data["message"])
                                yield TypeAdapter(ChunkedGenerationResponse).validate_python(data)
                            except GPTRouterStreamingError as e:
                                raise e
                            except Exception:
                                continue
                        return
            except (httpx.ReadTimeout, ServerError, httpx.ReadError) as err:
                logger.error(str(err))
                if attempt_number == retry_count:
                    logger.error("All retry attempts failed")
                    raise
                elif switch_models_order_on_timeout:
                    if attempt_number % 2:
                        self._reorder_models_on_timeout(payload)
                    logger.info(f"Retrying as per attempt #{attempt_number}")
            except httpx.HTTPStatusError as err:
                raise Exception(
                    f"HTTP Error {err.response.status_code}: {err.response.text}"
                ) from err
            except httpx.TimeoutException as err:
                raise TimeoutError("Request timed out") from err

    def _reorder_models_on_timeout(self, payload: dict) -> None:
        """Helper method to reorder models in payload when timeout occurs."""
        if not payload:
            payload = {}
        data = payload.get("data", [])
        try:
            if (
                len(data) >= 2
                and isinstance(data[0], dict)
                and isinstance(data[1], dict)
            ):
                data_0_order = data[0].get("order", None)
                data_1_order = data[1].get("order", None)
                if data_1_order and data_0_order:
                    (
                        payload["data"][0]["order"],
                        payload["data"][1]["order"],
                    ) = (
                        payload["data"][1]["order"],
                        payload["data"][0]["order"],
                    )
        except Exception as ex:
            logger.exception(f"Error reordering models: {ex}")

    def generate(
        self,
        *,
        ordered_generation_requests: List[ModelGenerationRequest],
        is_stream=False,
        model_router_metadata: Optional[GPTRouterMetadata] = None,
        **kwargs,
    ) -> Union[GenerationResponse, Generator[ChunkedGenerationResponse]]:
        api_path = "/v1/generate"
        api_method = "POST"
        api_payload = {
            "stream": is_stream,
            "data": [
                request.model_dump(exclude_none=True, by_alias=True)
                for request in ordered_generation_requests
            ],
        }
        api_payload = self.add_metadata_info(api_payload, model_router_metadata)
        if is_stream:
            return self.stream_events(
                path=api_path,
                method=api_method,
                payload=api_payload,
                **kwargs,
            )
        result = self._api_call(
            path=api_path,
            method=api_method,
            payload=api_payload,
            **kwargs,
        )
        return TypeAdapter(GenerationResponse).validate_python(result)

    async def agenerate(
        self,
        *,
        ordered_generation_requests: List[ModelGenerationRequest],
        is_stream=False,
        model_router_metadata: Optional[GPTRouterMetadata] = None,
        stop_stream_signal: Optional[asyncio.Event] = None,
        **kwargs,
    ) -> GenerationResponse:
        api_path = "/v1/generate"
        api_method = "POST"
        api_payload = {
            "stream": is_stream,
            "data": [
                request.model_dump(exclude_none=True, by_alias=True)
                for request in ordered_generation_requests
            ],
        }
        api_payload = self.add_metadata_info(api_payload, model_router_metadata)
        if is_stream:
            return self.astream_events(
                path=api_path,
                method=api_method,
                payload=api_payload,
                stop_stream_signal=stop_stream_signal,
                **kwargs,
            )
        result = await self._async_api_call(
            path=api_path,
            method=api_method,
            payload=api_payload,
            **kwargs,
        )
        return TypeAdapter(GenerationResponse).validate_python(result)

    async def agenerate_images(
        self, *, image_generation_request: ImageGenerationRequest
    ) -> List[ImageGenerationResponse]:
        api_path = "/v1/generate/generate-image"
        api_method = "POST"
        api_payload = image_generation_request.model_dump()

        api_response = await self._async_api_call(
            path=api_path,
            method=api_method,
            payload=api_payload,
        )
        generated_images = api_response.get("response", [])
        if isinstance(generated_images, dict):
            generated_images = self._extract_images_from_dict(generated_images)

        return [
            ImageGenerationResponse.model_validate(generated_img)
            for generated_img in generated_images
        ]

    def edit_image(self, *, image_edit_request: ImageEditRequest) -> List[ImageGenerationResponse]:
        """
        Edit an image based on the provided ImageEditRequest.
        This is a synchronous wrapper around the async method.
        
        Args:
            image_edit_request: The request containing the image and editing instructions
            
        Returns:
            List of image generation responses
        """
        import asyncio
        
        async def _async_call():
            return await self.aedit_image(image_edit_request=image_edit_request)
            
        return asyncio.run(_async_call())
    
    async def aedit_image(self, *, image_edit_request: ImageEditRequest) -> List[ImageGenerationResponse]:
        """
        Edit an image based on the provided ImageEditRequest.
        
        Args:
            image_edit_request: The request containing the image and editing instructions
            
        Returns:
            List of image generation responses
        """
        api_path = "/v1/generate/edit-image"
        api_method = "POST"
        api_payload = image_edit_request.model_dump()
        
        api_response = await self._async_api_call(
            path=api_path,
            method=api_method,
            payload=api_payload,
        )
        
        generated_images = api_response.get("response", {}).get("data", [])
        
        return [
            ImageGenerationResponse.model_validate({
                "url": img.get("url", None),
                "base64": img.get("b64_json", None),
                "finish_reason": "SUCCESS"
            })
            for img in generated_images
        ]

    def _extract_images_from_dict(self, image_dict: dict) -> List[dict]:
        if artifacts := image_dict.get("artifacts"):
            return artifacts

        if image_urls := image_dict.get("imageUrls"):
            return [{"url": url} for url in image_urls]

        return []



from pydantic import ConfigDict, BaseModel, Field
from enum import Enum
from typing import Dict, List, Literal, Optional, Any, Union
from humps import camelize


def to_camel(string):
    return camelize(string)


class GPTRouterMetadata(BaseModel):
    tag: Optional[str] = None
    history_id: Optional[str] = None
    created_by_user_id: Optional[str] = None
    appId: Optional[str] = None


class ThinkingType(str, Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"


class SearchContextSize(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class ExtendedThinking(BaseModel):
    type: ThinkingType
    budget_tokens: Optional[int] = 1024


class UserLocationApproximate(BaseModel):
    city: Optional[str] = None
    country: Optional[str] = None
    region: Optional[str] = None
    timezone: Optional[str] = None


class UserLocation(BaseModel):
    type: Optional[str] = None
    approximate: Optional[UserLocationApproximate] = None


class WebSearchOptions(BaseModel):
    search_context_size: Optional[SearchContextSize] = None
    user_location: Optional[Union[dict, UserLocation]] = None


class GenerationParams(BaseModel):
    messages: Optional[List[Any]] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    user: Optional[str] = None
    prompt: Optional[str] = None
    frequency_penalty: Optional[float] = None
    presence_penalty: Optional[float] = None
    functions: Optional[List[Any]] = None
    function_call: Optional[dict] = None
    tools: Optional[List[Any]] = None
    tool_choice: Optional[Union[str, dict]] = None
    response_format: Optional[Dict[str, Any]] = None
    parallel_tool_calls: Optional[bool] = None
    tag: Optional[str] = None
    user_id: Optional[str] = None
    history_id: Optional[str] = None
    appId: Optional[str] = None
    metadata: Optional[dict] = None
    raw_event: Optional[bool] = None
    thinking: Optional[ExtendedThinking] = None
    colasce_into_thinking_tag: Optional[bool] = None
    use_response_api: Optional[bool] = None
    web_search_options: Optional[WebSearchOptions] = None
    reasoning_effort: Optional[Literal["minimal", "low", "medium", "high"]] = None


class ModelGenerationRequest(BaseModel):
    model_name: str
    provider_name: str
    order: int = Field(int)
    prompt_params: Optional[GenerationParams] = Field(default={})
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True, protected_namespaces=())


class Usage(BaseModel):
    completion_tokens: Optional[int] = None
    prompt_tokens: Optional[int] = None
    total_tokens: Optional[int] = None


class Choice(BaseModel):
    index: int
    text: str
    finish_reason: str
    role: Optional[str] = None
    function_call: Optional[Any] = None
    citations: Optional[List[dict | Any]] = None


class GenerationResponse(BaseModel):
    id: str
    choices: List[Choice]
    model: str
    provider_id: Optional[str] = Field(None, alias="providerId")
    model_id: Optional[str] = Field(None, alias="modelId")
    meta: Optional[Usage] = None
    citations: Optional[List[str]] = None
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True, protected_namespaces=())


class ChunkedGenerationResponse(BaseModel):
    event: str
    data: dict
    provider: Optional[str] = None


class ImageGenerationRequest(BaseModel):
    model_name: str = Field(alias="model")
    provider_name: str = Field(alias="imageVendor")
    prompt: str
    num_images: int = Field(alias="numImages", default=1)
    width: Optional[int] = Field(None, alias="width")
    height: Optional[int] = Field(None, alias="height")
    num_inference_steps: Optional[int] = Field(None, alias="numInferenceSteps")
    app_id: Optional[str] = Field(None, alias="appId")
    user_id: Optional[str] = Field(None, alias="createdByUserId")
    history_id: Optional[str] = Field(None, alias="historyId")
    feature: Optional[str] = Field(None, alias="feature")
    model_config = ConfigDict(populate_by_name=True, protected_namespaces=())

    @property
    def size(self) -> Optional[str]:
        return f"{self.width}x{self.height}" if self.width and self.height else None

    def model_dump(self, *args, **kwargs):
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('by_alias', True)
        base_dict = super().model_dump(*args, **kwargs)
        if self.size is not None:
            base_dict["size"] = self.size
        return base_dict


class ImageGenerationResponse(BaseModel):
    url: str = None
    base64: str = None
    finish_reason: Optional[str] = Field(default="SUCCESS", alias="finishReason")


class ImageEditRequest(BaseModel):
    model_name: str = Field(alias="model", default="gpt-image-1")
    image: Union[str, List[str]] = Field(...)
    prompt: str = Field(...)
    mask: Optional[str] = Field(default=None)
    num_images: Optional[int] = Field(alias="n", default=1)
    quality: Optional[str] = Field(default="high")
    response_format: Optional[str] = Field(alias="responseFormat", default=None)
    size: Optional[str] = Field(default="1024x1024")
    user: Optional[str] = Field(default=None)
    model_config = ConfigDict(populate_by_name=True, protected_namespaces=())

    def model_dump(self, *args, **kwargs):
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('by_alias', True)
        base_dict = super().model_dump(*args, **kwargs)
        return base_dict


class ModelRouterGenerationRequest(BaseModel):
    model_name: str
    model_provider: str
    prompt_params: GenerationParams = Field(alias="promptParams")
    should_validate_token_count: Optional[bool] = Field(
        default=False, alias="shouldValidateTokenCount"
    )
    model_config = ConfigDict(populate_by_name=True, protected_namespaces=())

    def model_dump(self, *args, **kwargs):
        # Generate the standard dict first.
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('by_alias', True)
        result = super().model_dump(*args, **kwargs)

        result["providerName"] = self.model_provider
        result["modelName"] = self.model_name

        return result


class ModelRouterEmbeddingsGenerationParams(BaseModel):
    input: List[str]


class ModelRouterEmbeddingsGenerationRequest(ModelRouterGenerationRequest):
    prompt_params: ModelRouterEmbeddingsGenerationParams = Field(alias="promptParams")


class ModelRouterGenerationResponse(BaseModel):
    id: str
    choices: List[Choice]
    model: str
    provider_id: Optional[str] = Field(None, alias="providerId")
    model_id: Optional[str] = Field(None, alias="modelId")
    meta: Optional[Usage] = None
    model_config = ConfigDict(protected_namespaces=())


class EmbeddingsChoice(BaseModel):
    embedding: List[float]


class ModelRouterEmbeddingsGenerationResponse(BaseModel):
    choices: List[EmbeddingsChoice]
